#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

#include<dune/phasefield/localoperator/fem_base_chns.hh>


//Parameter class (for functions), FEM space and RangeField
template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
class FEM_FS_CHNS_U_Only :
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::NumericalJacobianVolume<FEM_FS_CHNS_U_Only<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
  public Dune::PDELab::NumericalJacobianApplyVolume<FEM_FS_CHNS_U_Only<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
  {
private:
  typedef typename CHContainer::field_type RF;
  typedef typename CHGFS::Traits::GridViewType::template Codim<0>::Entity   ENTITY;

  CHLocalBasisCache<CHGFS> chCache;
  VLocalBasisCache<NSGFS> vCache;

  typedef GFS_LocalViewer<CHGFS, CHContainer, ENTITY, RF> CHLView;
  mutable CHLView chOldLocal;

  typedef GFS_LocalViewer<NSGFS, NSContainer, ENTITY, RF> NSLView;
  mutable NSLView nsLocal;


  Param& param; // parameter functions

public:
  enum {doPatternVolume=true};
  enum {doAlphaVolume=true};

  //constructor
  FEM_FS_CHNS_U_Only(Param& param_, CHGFS& chGfs_, CHContainer& chVOld_, NSGFS& nsGfs_, NSContainer& nsV) :
    chOldLocal(chGfs_,chVOld_),
    nsLocal(nsGfs_,nsV),
    param(param_)
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    //Quadrature Rule
    const int dim = EG::Entity::dimension;
    const int order = 5; //TODO: Choose more educated
    auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);

    chOldLocal.DoEvaluation(eg);
    //chOldLocal.DoEvaluation(eg);
    nsLocal.DoSubEvaluation(eg,child(nsLocal.lfs,Dune::Indices::_0));

    for(const auto& ip : rule)
    {
      const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
      auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());

      CHVariables_U_PFOld_Extension<RF,dim,LFSV> ch(chCache, ip.position(), lfsv, x.base(),chOldLocal.lv, S);
      NSVariables<RF,dim,decltype(nsLocal.lfs)> ns(vCache, ip.position(), nsLocal.lfs, nsLocal.lv, S);

      RF pf_c = ch.pf;
      RF pf_f = ch.pf + 2*param.delta*(1-ch.pf);
      Dune::FieldVector<RF,dim> gradpf_f(0.0);
      gradpf_f.axpy((1-2*param.delta),ch.gradpf);

      RF pf_c_Old = ch.pfOld;
      RF pf_c_reg = pf_c+param.delta;
      RF pf_c_reg_Old = pf_c_Old + param.delta;

      for (size_t i=0; i<ch.pfspace.size(); i++)
      {
        RF Rspeed = param.dw.eval_q((ch.pfOld-0.05)/0.7)/0.7;
        RF R1 = -Rspeed/param.eps*
                  (param.reaction.eval(ch.u)+param.phys.curvatureAlpha*ch.xi)*ch.basis.phi[i]; //TODO Scale with sigma13

      // integrate: (pf-pfold)*phi_i/delta_t + gradpf * grad phi_i + ...
        r.accumulate(ch.pfspace,i,factor*(
          (ch.pf-ch.pfOld)*ch.basis.phi[i]/param.time.dt
        ));

        r.accumulate(ch.xispace,i,factor*(
          (ch.xi-0)*ch.basis.phi[i]
        ));


      // integrate (pf-pfold)*(u-uast)*phi_i/delta_t + pfold*(u-uold)*phi_i/delta_t + D*pf*gradu*grad phi_i = 0
        r.accumulate(ch.uspace,i,factor*(

          //(u-param.uAst)*phi[i]/param.dt //For Testing

          R1 *(ch.u-param.phys.uAst)
          + pf_c_reg_Old*(ch.u-ch.uOld)*ch.basis.phi[i]/param.time.dt
          + param.phys.D*pf_c_reg_Old*(ch.gradu*ch.basis.gradphi[i][0])
          + ( pf_f*(ns.v*ch.gradu) )*ch.basis.phi[i]
          - pf_c_reg*param.reaction.eval_bulkreaction()*ch.basis.phi[i]//Bulk reaction
          ));
      }
    }
  }
};
