#pragma once


#include<dune/pdelab/constraints/common/constraints.hh>
#include<dune/pdelab/constraints/conforming.hh>

#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include<dune/pdelab/gridfunctionspace/subspace.hh>
#include <dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh>

//_________________________________________________________________________________________________________________________


template<typename RF, typename ES, typename V_FEM, typename PHI_FEM>
class GFS_1p_chns
{
public:
  typedef Dune::PDELab::ConformingDirichletConstraints ConstraintsAssembler;
  typedef Dune::PDELab::ISTL::VectorBackend<> VectorBackend;
  typedef Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::fixed> VB_Block;
  typedef Dune::PDELab::LexicographicOrderingTag OrderingTagL;
  typedef Dune::PDELab::EntityBlockedOrderingTag OrderingTagB;

  ConstraintsAssembler conphi, conmu;

  typedef Dune::PDELab::GridFunctionSpace<ES, V_FEM, ConstraintsAssembler, VectorBackend> VX_GFS;
  VX_GFS vxGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, PHI_FEM, ConstraintsAssembler, VectorBackend> PHI_GFS;
  PHI_GFS phiGfs;

  typedef Dune::PDELab::GridFunctionSpace<ES, PHI_FEM, ConstraintsAssembler, VectorBackend> MU_GFS;
  MU_GFS muGfs;

  typedef Dune::PDELab::CompositeGridFunctionSpace<VectorBackend, OrderingTagL, VX_GFS> NS;
  NS ns;

  typedef Dune::PDELab::CompositeGridFunctionSpace<VB_Block, OrderingTagB, PHI_GFS, MU_GFS> CH;
  CH ch;

  typedef typename NS::template ConstraintsContainer<RF>::Type NS_CC;
  NS_CC nsCC;

  typedef typename CH::template ConstraintsContainer<RF>::Type CH_CC;
  CH_CC chCC;

  GFS_1p_chns(const ES& es, const V_FEM& vFem, const PHI_FEM& phiFem) :
      conphi(),conmu(),
      vxGfs(es,vFem),
      phiGfs(es,phiFem, conphi),
      muGfs(es,phiFem, conmu),
      ns(vxGfs),
      ch(phiGfs, muGfs),
      nsCC(),
      chCC()
  {
    vxGfs.name("velocity");
    phiGfs.name("phasefield");
    muGfs.name("potential");
  }

  template<typename Bdry>
  void updateConstraintsContainer(const Bdry& bdry)
  {
    Dune::PDELab::constraints(bdry.bdryNS,ns,nsCC, 0);
    Dune::PDELab::constraints(bdry.bdryCH,ch,chCC, 0);
  }
};

//_________________________________________________________________________________________________________________________


template<typename GFS, typename NS_V, typename CH_V>
class DGF_1p_chns
{
public:

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::NS,Dune::TypeTree::TreePath<0> > VXSubGFS;
  VXSubGFS vxSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<VXSubGFS,NS_V> VXDGF;
  VXDGF vxDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<0> > PhiSubGFS;
  PhiSubGFS phiSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<PhiSubGFS,CH_V> PhiDGF;
  PhiDGF phiDgf;

  typedef Dune::PDELab::GridFunctionSubSpace<typename GFS::CH,Dune::TypeTree::TreePath<1> > MuSubGFS;
  MuSubGFS muSubGfs;
  typedef Dune::PDELab::DiscreteGridFunction<MuSubGFS,CH_V> MuDGF;
  MuDGF muDgf;


  DGF_1p_chns(const GFS& gfs, const NS_V& nsV, const CH_V& chV) :
      vxSubGfs(gfs.ns),
      vxDgf(vxSubGfs,nsV),
      phiSubGfs(gfs.ch),
      phiDgf(phiSubGfs,chV),
      muSubGfs(gfs.ch),
      muDgf(muSubGfs,chV)
  {
  }

  template<typename VTKWRITER>
  void addToVTKWriter(VTKWRITER& vtkwriter)
  {
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<VXDGF> >(vxDgf,"vx"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<PhiDGF> >(phiDgf,"phi"));
    vtkwriter.addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<MuDGF> >(muDgf,"mu"));
  }
};

//_________________________________________________________________________________________________________________________

template<typename IniVX, typename IniPhi, typename IniMu>
class Ini_1p_chns
{
public:
  IniVX iniVx;
  IniPhi iniPhi;
  IniMu iniMu;
  typedef Dune::PDELab::CompositeGridFunction<IniVX> IniNS;
  typedef Dune::PDELab::CompositeGridFunction<IniPhi,IniMu> IniCH;
  IniNS iniNS;
  IniCH iniCH;

  template<typename GV, typename Params>
  Ini_1p_chns(const GV& gv, const Params& params) :
      iniVx(gv, params),
      iniPhi(gv, params),
      iniMu(gv, params),
      iniNS(iniVx),
      iniCH(iniPhi,iniMu)
  {
  }
};

//_________________________________________________________________________________________________________________________

template<typename BdryVX,  typename BdryPhi, typename BdryMu>
class Bdry_1p_chns
{
public:
  BdryVX bdryVx;
  BdryPhi bdryPhi;
  BdryMu bdryMu;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryVX> BdryNS;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryPhi, BdryMu> BdryCH;
  BdryNS bdryNS;
  BdryCH bdryCH;

  Bdry_1p_chns() :
      bdryVx(),
      bdryPhi(),
      bdryMu(),
      bdryNS(bdryVx),
      bdryCH(bdryPhi,bdryMu)
  {
  }
};

template<typename BdryVX,  typename BdryPhi, typename BdryMu>
class Bdry_1p_chns_Explicit_Domain
{
public:
  BdryVX bdryVx;
  BdryPhi bdryPhi;
  BdryMu bdryMu;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryVX> BdryNS;
  typedef Dune::PDELab::CompositeConstraintsParameters<BdryPhi, BdryMu> BdryCH;
  BdryNS bdryNS;
  BdryCH bdryCH;

  template<typename DomainSize>
  Bdry_1p_chns_Explicit_Domain(DomainSize ds) :
      bdryVx(ds),
      bdryPhi(ds),
      bdryMu(ds),
      bdryNS(bdryVx),
      bdryCH(bdryPhi,bdryMu)
  {
  }
};
